#include <DRV8835MotorShield.h>
#include "SerialCommand.h"
#include "EEPROMAnything.h"

#define SerialPort Serial1
#define Baudrate 9600

/*
#define MotorRightForward 10
#define MotorRightBackward 6
#define MotorLeftForward 9
#define MotorLeftBackward 5
*/
DRV8835MotorShield motors;

int speedLeft;
int speedRight;

SerialCommand sCmd(SerialPort);
bool debug = false;
bool run;
unsigned long previous, calculationTime;
double iTerm, lastErr;

const int sensor[] = {A5, A4, A3, A2, A1, A0};

struct param_t 
{
  unsigned long cycleTime;

  int black[6];
  int white[6];
  int power;
  float diff;
  float kp;
  float ki;
  float kd;
} params;

int normalised[6];
float debugPosition;

void setup()
{
  //analogReference(DEFAULT);
  
  SerialPort.begin(Baudrate);
  sCmd.addCommand("set", onSet);
  sCmd.addCommand("debug", onDebug);
  sCmd.addCommand("calibrate", onCalibrate);
  sCmd.addCommand("start", onRun);
  sCmd.addCommand("reset", onReset);
  sCmd.addCommand("stop", onStop);
  sCmd.setDefaultHandler(onUnknownCommand);

  EEPROM_readAnything(0, params);

  SerialPort.println("ready");
}

void loop()
{
  sCmd.readSerial();
  unsigned long current = micros();
  if (current - previous >= params.cycleTime)
  {
    previous = current;
    
    /*code die cyclisch moet uitgelezen worden programmeer je hier*/
    for (int i = 0; i < 6; i++) normalised[i] = map(analogRead(sensor[i]),params.black[i], params.white[i], 0, 1000);

    int index = 0;
    float position;
    for (int i = 1; i < 6; i++) if (normalised[i] < normalised[index]) index = i;

    if (normalised[index] > 850) run = false;
    
    if (index == 0) position = -30;
    else if (index == 5) position = +30;
    else {
      int sNul = normalised[index];
      int sMinEen = normalised[index-1];
      int sPlusEen = normalised[index+1];

      float b = sPlusEen - sMinEen;
      b = b / 2;

      float a = sPlusEen - b - sNul;

      position = -b / (2 * a);
      position += index;
      position -= 2.5;

      position *= 15;
    }
    debugPosition = position;

    //P-regelaar 
    float error = -position;
    float output = error * params.kp;

    //I-regelaar
    iTerm += params.ki * error;
    iTerm = constrain(iTerm, -510, 510);
    output += iTerm;

    //D-regelaar
    output += params.kd * (error - lastErr);
    lastErr = error;

    //Output begrenzen 
    output = constrain(output, -510, 510);

    int powerLeft = 0;
    int powerRight = 0;

    if (run) if (output >= 0) {
      powerLeft = constrain(params.power + params.diff * output, -255, 255);
      powerRight = constrain(powerLeft - output, -255, 255);
      powerLeft = powerRight + output;
    }
    else {
      powerRight = constrain(params.power - params.diff * output, -255, 255);
      powerLeft = constrain(powerRight + output, -255, 255);
      powerRight = powerLeft - output;
    }

    /*analogWrite(MotorLeftForward, powerLeft > 0 ? powerLeft : 0);
    analogWrite(MotorLeftBackward, powerLeft < 0 ? -powerLeft : 0);
    analogWrite(MotorRightForward, powerRight > 0 ? powerRight : 0);
    analogWrite(MotorRightBackward, powerRight < 0 ? -powerRight : 0);*/
    /*if (powerLeft > 0) {
      powerLeft = powerLeft;
    }
    else if (powerLeft < 0) {
      powerLeft = 0 - powerLeft;
    }

    if (powerRight > 0) {
      powerRight = powerRight;
    }
    else if (powerRight < 0) {
      powerRight = 0 - powerRight;
    }*/

    /*Serial.print("PowerRight: ");
    Serial.println(powerRight);
    Serial.print("PowerLeft: ");
    Serial.println(powerLeft);*/
    
    motors.setM1Speed(powerLeft);
    motors.setM2Speed(powerRight);
  }

  unsigned long difference = micros() - current;
  if (difference > calculationTime) calculationTime = difference;
} 

void onSet()
{
  char* param = sCmd.next();
  char* value = sCmd.next();

  if (strcmp(param, "cycle") == 0) {
    long newCycleTime = atol(value);
    float ratio = ((float) newCycleTime) / ((float) params.cycleTime);

    params.ki *= ratio;
    params.kd /= ratio;

    params.cycleTime = newCycleTime;
  }

  else if (strcmp(param, "power") == 0) params.power = atol(value);
  else if (strcmp(param, "diff") == 0) params.diff = atof(value);
  else if (strcmp(param, "kp") == 0) params.kp = atof(value); 
  else if (strcmp(param, "ki") == 0) {
    float sampleTimeInSec = ((float) params.cycleTime) / 1000000;
    params.ki = atof(value) * sampleTimeInSec;
  }
  else if (strcmp(param, "kd") == 0) {
    float cycleTimeInSec = ((float) params.cycleTime) / 1000000;
    params.kd = atof(value) / cycleTimeInSec;
  }

  EEPROM_writeAnything(0, params);
}

void onDebug() {
  SerialPort.print("cycle time: ");
  SerialPort.println(params.cycleTime);

  /*parameters weergeven met behulp van het debug commande doe je hier*/
  SerialPort.print("black: ");
  for (int i = 0; i < 6; i++) {
    SerialPort.print(params.black[i]);
    SerialPort.print(" ");
  }
  SerialPort.println(" ");

  SerialPort.print("white: ");
  for (int i = 0; i < 6; i++) {
    SerialPort.print(params.white[i]);
    SerialPort.print(" ");
  }
  SerialPort.println(" ");

  SerialPort.print("normalised: ");
  for (int i = 0; i < 6; i++) {
    SerialPort.print(normalised[i]);
    SerialPort.print(" ");
  }
  SerialPort.println(" ");

  SerialPort.print("position: ");
  SerialPort.println(debugPosition);

  SerialPort.print("power: ");
  SerialPort.println(params.power);
  SerialPort.print("diff: ");
  SerialPort.println(params.diff);
  SerialPort.print("kp: ");
  SerialPort.println(params.kp);

  float cycleTimeInSec = ((float) params.cycleTime) / 1000000;
  float ki  = params.ki / cycleTimeInSec;
  SerialPort.print("Ki: ");
  SerialPort.println(ki);

  float kd  = params.kd * cycleTimeInSec;
  SerialPort.print("Kd: ");
  SerialPort.println(kd);

  SerialPort.print("calculation time: ");
  SerialPort.println(calculationTime);
  calculationTime = 0;
}

void onUnknownCommand(char* command)
{
  SerialPort.print("Unknown Command: \"");
  SerialPort.print(command);
  SerialPort.println("\"");
}

void onCalibrate() {
  char* param = sCmd.next();

  if (strcmp(param, "black") == 0) {
    SerialPort.print("start calibrating black... ");
    for (int i = 0; i < 6; i++) params.black[i] = analogRead(sensor[i]);
    SerialPort.println("done");
  }
  else if (strcmp(param, "white") ==0) {
    SerialPort.print("start calibrating white... ");
    for (int i = 0; i < 6; i++) params.white[i] = analogRead(sensor[i]);
    SerialPort.println("done");
  }

  EEPROM_writeAnything(0, params);
}

void onRun() {
  if (SerialPort.read() == '1') {
    run = false;
    /*analogWrite(MotorRightForward, 0);
    analogWrite(MotorRightBackward, 0);
    analogWrite(MotorLeftForward, 0);
    analogWrite(MotorLeftBackward, 0);*/
    /*speedRight = 0;
    speedLeft = 0;
    motors.setM1Speed(speedRight);
    motors.setM2Speed(speedRight);*/
    
    SerialPort.println("Stopped");
  }
  else {
    run = true;
    SerialPort.println("Started");
  }
}

void onStop() {
  run = false;
  SerialPort.println("Stopped");
}

void onReset() {
    params.kp = 11.5;
    params.ki = 0;
    iTerm = 0;
    params.kd = 0;
    lastErr = 0;
    params.diff = 0.57;
    params.power = 100;
    params.cycleTime = 500;

    EEPROM_writeAnything(0, params);
    SerialPort.println("Reset done");
}